import asyncio
import logging
import os
import signal
import subprocess
import sys
from enum import Enum
from pathlib import Path
from typing import Any, Callable, TypeVar, override

import rich.console
import rich.traceback

from ai_parade._cli.ai_parade_puppet.communication import (
	ConnectionClosed,
	receive_async,
	send_async,
)
from ai_parade._toolkit.constants import TRACE_LVL, VENV_DIR
from ai_parade._toolkit.data.output import ModelOutput
from ai_parade._toolkit.datasets.Dataset import Dataset
from ai_parade._toolkit.enums import AIRuntime
from ai_parade._toolkit.metadata.models.api import RunnerGetter
from ai_parade._toolkit.metadata.models.final import ModelMetadata

from .ModelRunner import LoadedModelRunner, ModelRunner, TReturn

logger = logging.getLogger(__name__)

ReturnType = TypeVar("ReturnType")


class MessageKind(Enum):
	"""Message kinds for the remote runner."""

	OK = "OK"
	"""Success message."""

	ERROR = "ERROR"
	"""Error message."""


class RemoteRunner(
	ModelRunner["RemoteRunner", None], LoadedModelRunner["RemoteRunner", None]
):
	"""Runs the model as separate process."""

	MAX_MEM = 6_500_000_000

	def __init__(
		self,
		model_metadata: ModelMetadata,
		runtime: AIRuntime,
		weights: Path,
		local_runner: RunnerGetter,
	):
		"""Initialize the remote runner, and **NOT** yet start the process

		Args:
			model_metadata: metadata
			model_format: Format of the model.
			weights: Path to the model weights
			local_runner: runner used in the target or instead of the remote runner in case we are running runners locally

		To start the process, use `with remote_runner_instance:`
		or call directly`RemoteRunner.__enter__()`
		"""
		super().__init__(model_metadata, weights)
		assert model_metadata.repository_path is not None

		self.local_runner = local_runner
		self.venv_path = model_metadata.repository_path / VENV_DIR
		self.runtime = runtime
		self._response_future = None

		self._writer = None
		self._isConnected = None
		"""Future that resolves when the remote runner is connected or None if not connected"""

		self._debugging = False
		try:
			import debugpy  # type: ignore

			self._debugging = debugpy.is_client_connected()

		except ImportError:
			pass

	async def _start_remote_runner(self):
		self._server = await asyncio.start_server(
			self._handle_client,
			host="localhost",
			port=0,  # choose a random available port
		)
		port = self._server.sockets[0].getsockname()[1]

		self._server_task = asyncio.create_task(self._server.serve_forever())
		self._start_puppet(port)

		self._isConnected = asyncio.Future()

		logger.debug(f"Waiting for the remote runner to connect on port {port}...")
		await asyncio.wait_for(
			self._isConnected, timeout=300 if self._debugging else 60
		)
		logger.debug("Remote runner connected")

		# copy current value to the remote - triggers send
		await self.set_output_directory(self.output_directory)

	@override
	async def __aenter__(self):
		await self._start_remote_runner()
		try:
			await self._send_command(self.__aenter__.__name__)
		except Exception as e:
			await self.__aexit__(e)
			raise

		return self

	@override
	async def __aexit__(self, *err: Any):
		logger.debug("Exiting remote runner")

		self._isConnected = None

		assert self._writer is not None
		# we must close the writer before the server
		self._writer.close()
		self._writer = None
		# self._server.close()
		self._server_task.cancel()
		await self._server.wait_closed()

		self._process.kill()

	@override
	async def set_output_directory(self, value: Path | None):
		await super().set_output_directory(value)
		await self._send_command(self.set_output_directory.__name__, value)

	def ensure_entered(self):
		if self._isConnected is None or not self._isConnected.done():
			raise RuntimeError("Runner used without `with` statement")

	@override
	async def inference(self, dataset: Dataset) -> list[ModelOutput]:
		self.ensure_entered()
		return await self._send_command(self.inference.__name__, dataset)

	@override
	async def run(self, what: Callable[["RemoteRunner"], TReturn], *args) -> TReturn:
		self.ensure_entered()
		return await self._send_command(self.run.__name__, what, *args)

	@override
	async def get_model_size(self) -> int | None:
		size = None
		if self._isConnected is None or not self._isConnected.done():
			try:
				await self._start_remote_runner()
				size = await self._send_command(self.get_model_size.__name__)
			finally:
				await self.__aexit__(*sys.exc_info())
			return size
		return await self._send_command(self.get_model_size.__name__)

	@property
	@override
	def model(self):
		raise RuntimeError("Trying to get the model from a remote runner")

	def _start_puppet(self, port: int):
		environment = os.environ.copy()
		environment.pop("MPLBACKEND", None)  # Do not share the matplotlib backend
		environment["PATH"] = (
			str(self.venv_path / "bin") + os.pathsep + environment["PATH"]
		)

		logger.info("Starting remote runner process")
		cli_parameters = [
			str(self.venv_path / "bin" / "ai-parade-puppet"),
			"--weights",
			str(self.weights),
			"--repository-path",
			str(self._model_metadata.repository_path),
			"--model-name",
			self._model_metadata.name,
			"--runtime",
			self.runtime.value,
			"--port",
			str(port),
			"--max-mem",
			str(RemoteRunner.MAX_MEM),
		] + (["--debugger-port", str(5678)] if self._debugging else [])
		logger.log(TRACE_LVL, f"Calling ai-parade-puppet with {cli_parameters}")
		self._process = subprocess.Popen(
			cli_parameters,
			env=environment,
		)

	async def _send_command(self, attribute: str, *params: Any):
		assert self._writer is not None
		logger.debug("Sending command: " + attribute)
		logger.log(TRACE_LVL, f"Command {attribute}({params})")

		if self._response_future is not None:
			raise RuntimeError("Another command is already being processed")
		self._response_future = asyncio.Future()
		data = (attribute, params)

		try:
			if self._debugging:
				await asyncio.wait_for(send_async(self._writer, data), timeout=60)
				result = await self._response_future
			else:
				await asyncio.wait_for(send_async(self._writer, data), timeout=10)
				result = await asyncio.wait_for(
					self._response_future, timeout=15 * 60 if self._debugging else None
				)
		except ConnectionClosed as e:
			logger.debug("Connection closed trying to get more info:")
			return_code = self._process.poll()
			if return_code is not None and -return_code in [
				signal.SIGKILL,
				signal.SIGTERM,
			]:  # negative return code == signal number on posix
				logger.info(
					f"Remote runner process has been killed (return code {return_code}), we assume it's because of OOM."
				)
				raise MemoryError(e)
			logger.info(
				f"Remote runner process has died with return code {return_code}"
			)
			raise e
		finally:
			self._response_future = None

		logger.debug(f"Command {attribute} done.")
		return result

	async def _handle_client(
		self, reader: asyncio.StreamReader, writer: asyncio.StreamWriter
	):
		assert self._isConnected is not None and not self._isConnected.done()
		self._isConnected.set_result(True)
		assert self._writer is None
		self._writer = writer

		while True:
			try:
				response = await receive_async(reader)
				self._parse_response(response)

			except Exception as e:
				if self._isConnected is None:
					break
				logger.error(f"Receive results in exception: {e}")
				if self._response_future is not None:
					self._response_future.set_exception(e)
				break

	def _parse_response(self, received: tuple[MessageKind, Any]):
		kind, output = received

		logger.log(
			TRACE_LVL, f"Received data of kind: {kind.value} with output: {output}"
		)
		if self._response_future is None:
			return

		match kind:
			case MessageKind.OK:
				self._response_future.set_result(output)
			case MessageKind.ERROR:
				inner_error = output[0]
				# inner_error.add_note(output[1])
				if "memory" in str(inner_error).lower():
					logger.info("Got error containing keyword 'memory', assume OOM")
					e = MemoryError(inner_error)
				else:
					e = RemoteError(
						output[1],
						"Error in the remote process. \n" + str(inner_error),
					)
				e.__cause__ = inner_error
				self._response_future.set_exception(e)


class RemoteError(Exception):
	"""
	Exception raised when the remote runner returns an error.
	"""

	def __init__(self, rich_traceback: rich.traceback.Traceback, *args: object) -> None:
		self.rich_traceback = rich_traceback
		super().__init__(*args)

	def __str__(self) -> str:
		console = rich.console.Console()
		with console.capture() as capture:
			console.print(self.rich_traceback)

		return "Remote error:\n" + capture.get()



import argparse
import asyncio
import inspect
import logging
import resource
import socket
import sys
from pathlib import Path

from rich.traceback import Traceback

import ai_parade._compat  # noqa: F401
from ai_parade._toolkit.run.RemoteRunner import MessageKind
from ai_parade._toolkit.run.runner_selector import get_runner
from ai_parade.toolkit import (
	AIRuntime,
	CaptureOutputs,
	ModelMetadata,
	get_local_metadata,
)

from .communication import ConnectionClosed, receive_async, send

logger = logging.getLogger(__name__)

# log everything! (then save it)
logging.root.setLevel(0)


def prepare_args(
	args: argparse.Namespace,
) -> tuple[AIRuntime, ModelMetadata, Path]:
	metadata_list = asyncio.run(get_local_metadata(args.repository_path))
	metadata = list(filter(lambda m: m.name == args.model_name, metadata_list))[0]
	runtime = AIRuntime(args.runtime)

	return runtime, metadata, Path(args.weights)


def define_params():
	parser = argparse.ArgumentParser(
		description="CLI tool for inference and data conversion"
	)
	parser.add_argument(
		"-e",
		"--runtime",
		type=str,
		required=True,
		help="Desired AI runtime of the model",
	)
	parser.add_argument(
		"-w", "--weights", type=str, required=True, help="Path to the model weights"
	)
	parser.add_argument(
		"-n",
		"--model-name",
		type=str,
		required=True,
		help="Name of model (as in the name in its metadata file)",
	)
	parser.add_argument(
		"-r",
		"--repository-path",
		type=Path,
		required=True,
		help="Name of the model repository",
	)

	parser.add_argument(
		"-p", "--port", type=int, required=True, help="Communication port"
	)

	parser.add_argument(
		"--max-memory",
		type=int,
		default=2**30,
		help="Maximum amount of memory to use in bytes",
	)

	parser.add_argument(
		"--debugger-port",
		type=int,
		help="Enable debug with debugger port",
	)
	parser.add_argument(
		"--debugger-host",
		type=str,
		help="Set debugger host (default: localhost)",
		default="localhost",
	)

	args = parser.parse_args()
	is_debugging = False
	if args.debugger_port is not None:
		is_debugging = True
		try:
			import debugpy  # type: ignore

			print(f"Is debugging? {debugpy.is_client_connected()}")

			debugpy.listen((args.debugger_host, args.debugger_port))
			print(f"Waiting for debugger to connect at port {args.debugger_port}...")
			debugpy.wait_for_client()
			print("Debugger connected.")
		except ImportError:
			print(
				"Debugpy is not installed. Please install it to enable debugging of RemoteRunner."
			)
	runtime, metadata, weights = prepare_args(args)

	resource.setrlimit(resource.RLIMIT_DATA, (args.max_memory, args.max_memory))

	asyncio.run(run_as_puppet(args, runtime, metadata, weights, is_debugging))


def main():
	define_params()


if __name__ == "__main__":
	main()


async def run_as_puppet(
	args: argparse.Namespace,
	runtime: AIRuntime,
	metadata: ModelMetadata,
	weights: Path,
	is_debugging: bool,
):
	"""
	Delegates the command from the main process to the runner.

	Args:
		args: Parsed CLI arguments.
		runtime: Desired runtime of the model.
		metadata: Metadata of the model to run.
		weights: Path to the model weights
	"""

	with socket.socket(socket.AF_INET, socket.SOCK_STREAM) as connection:
		connection.connect(("localhost", args.port))
		runner = get_runner(metadata, weights, runtime=runtime, use_remote=False)

		while True:
			try:
				return_value = None

				data = await receive_async(connection)
				attribute, params = data

				if not hasattr(runner, attribute):
					raise RuntimeError(f"Unknown attribute `{attribute}`")
				maybe_method = getattr(runner, attribute)
				if inspect.ismethod(maybe_method):
					with CaptureOutputs(
						save_dir=runner.output_directory, catch_errors=False
					):
						return_value = maybe_method(*params)
						if inspect.iscoroutine(return_value):
							return_value = await return_value

					if attribute == "__enter__" or attribute == "__aenter__":
						runner = return_value
						# __enter__ outputs the loaded runner, which is not picklable
						# so just don't send it, after all the remote runner should do is to return itself
						return_value = None
				else:
					if params:
						assert len(params) == 1
						return_value = setattr(runner, attribute, params[0])
					else:
						return_value = getattr(runner, attribute)

				send(connection, (MessageKind.OK, return_value))
			except ConnectionClosed as e:
				if is_debugging:
					print("[Puppet] Connection closed:")
					print(e)
				logger.debug(f"Connection closed: {e}")
			except Exception as e:
				if is_debugging:
					print("[Puppet] Error:")
					print(e)
				rich_traceback = Traceback.from_exception(*sys.exc_info())

				# Custom errors can't be unpickled, so we don't send them
				if e.__class__.__module__ == "builtins":
					error_message = (e, rich_traceback)
				else:
					error_message = (RuntimeError(*e.args), rich_traceback)

				send(connection, (MessageKind.ERROR, error_message))



from pathlib import Path

from typer.testing import CliRunner

from ai_parade._cli.ai_parade.main import app

from ._common import setup_cwd_any

runner = CliRunner()


class TestCheck:
	def test_no_requirements(self, tmp_path, monkeypatch):
		setup_cwd_any(
			tmp_path,
			monkeypatch,
			repo_path=Path("fixtures/FooBarModel"),
			skip=["requirements.txt"],
		)
		result = runner.invoke(app, "check")
		assert result.exit_code == 8

	def test_no_requirements_pin(self, tmp_path, monkeypatch):
		setup_cwd_any(
			tmp_path,
			monkeypatch,
			repo_path=Path("fixtures/FooBarModel"),
			skip=["requirements.txt"],
		)
		assert not Path("requirements.txt").exists()

		result = runner.invoke(app, "check --pin-installed")

		assert Path("requirements.txt").exists()
		assert result.exit_code == 0

	def test_no_venv(self, tmp_path, monkeypatch):
		setup_cwd_any(
			tmp_path,
			monkeypatch,
			repo_path=Path("fixtures/FooBarModel"),
			skip=[".venv"],
		)

		result = runner.invoke(app, "check")

		assert result.exit_code == 8

	def test_ok(self, tmp_path, monkeypatch):
		setup_cwd_any(
			tmp_path,
			monkeypatch,
			repo_path=Path("fixtures/FooBarModel"),
		)

		result = runner.invoke(app, "check")
		print(result.output)

		assert result.exit_code == 0

	def test_with_inference(self, tmp_path, monkeypatch):
		setup_cwd_any(
			tmp_path,
			monkeypatch,
			repo_path=Path("fixtures/FooBarModel"),
		)

		result = runner.invoke(app, "check FooBarModel weights.pth")
		print(result.output)

		assert result.exit_code == 0

	def test_with_inference_missing_weights(self, tmp_path, monkeypatch):
		setup_cwd_any(
			tmp_path,
			monkeypatch,
			repo_path=Path("fixtures/FooBarModel"),
			skip=["weights.pth"],
		)

		result = runner.invoke(app, "check FooBarModel weights.pth")
		print(result.output)

		assert result.exit_code != 0



from functools import partial
from math import ceil, floor

import cv2 as cv
import numpy as np

from .ImageInput import ImageInput


def resize(
	input: np.ndarray,
	target_size: tuple[int, int],
	padding: None | tuple[float, float, float] = None,
) -> np.ndarray:
	assert len(input.shape) == 3
	assert input.shape[2] == 3
	if padding is not None:
		# mind that x = height, y = width
		x_rescale = target_size[0] / input.shape[0]
		y_rescale = target_size[1] / input.shape[1]
		rescale = min(x_rescale, y_rescale)
		input = cv.resize(
			input, (int(input.shape[1] * rescale), int(input.shape[0] * rescale))
		)
		x_padding = (target_size[0] - input.shape[0]) / 2
		y_padding = (target_size[1] - input.shape[1]) / 2
		input = cv.copyMakeBorder(
			input,
			floor(x_padding),
			ceil(x_padding),
			floor(y_padding),
			ceil(y_padding),
			cv.BORDER_CONSTANT,
			value=padding,
		)
		return input
	else:
		return cv.resize(input, target_size)


def preprocess(mat: np.ndarray, image_input: ImageInput) -> np.ndarray:
	assert len(mat.shape) == 3
	assert mat.shape[2] == 3
	# RGB -> imageInput.channelOrder
	match image_input.channelOrder:
		case ImageInput.ChannelOrder.RGB:
			pass
		case ImageInput.ChannelOrder.RGBA:
			mat = cv.cvtColor(mat, cv.COLOR_RGBA2RGB)
		case ImageInput.ChannelOrder.BGR:
			mat = cv.cvtColor(mat, cv.COLOR_BGR2RGB)
		case ImageInput.ChannelOrder.BGRA:
			mat = cv.cvtColor(mat, cv.COLOR_BGRA2RGB)

	# Resize
	mat_resized = resize(mat, (image_input.width, image_input.height))

	# Convert UINT8 to data type imageInput.dataType
	match image_input.dataType:
		case ImageInput.DataType.float32:
			mat = mat_resized.astype(np.float32) / 255
		case ImageInput.DataType.uint8:
			pass

	# Normalize if available
	if image_input.means is not None:
		assert image_input.channels == 3
		mean = np.array(image_input.means, dtype=np.float32)
		mat = cv.subtract(mat, mean)

	if image_input.stds is not None:
		assert image_input.channels == 3
		std = np.array(image_input.stds, dtype=np.float32)
		mat = cv.divide(mat, std)

	# Transpose to the imageInput.dataOrder
	if image_input.dataOrder == ImageInput.DataOrder.NCHW:
		mat = np.transpose(mat, (2, 0, 1))

	return mat


def preprocess_fx(image_input: ImageInput):
	return partial(preprocess, image_input=image_input)



from ..enums import (
	AIRuntime,
	HardwareAcceleration,
	ModelFormat,
	Platform,
	Quantization,
)
from ._common import ModelAttributes

runtime = AIRuntime.LiteRT
capabilities = {
	Platform.Android: {
		Quantization.none: {
			HardwareAcceleration.XNNPACK,
			HardwareAcceleration.GPU,  # == OpenCL or EGL
			HardwareAcceleration.QNN,
			HardwareAcceleration.Mediatek,
			HardwareAcceleration.GoogleTensor,
		},
		Quantization.int8: {
			HardwareAcceleration.XNNPACK,
			HardwareAcceleration.GPU,  # == OpenCL or EGL
			HardwareAcceleration.QNN,
			HardwareAcceleration.Mediatek,
			HardwareAcceleration.GoogleTensor,
		},
		Quantization.float16: {
			HardwareAcceleration.XNNPACK,
			HardwareAcceleration.GPU,  # == OpenCL or EGL
			HardwareAcceleration.QNN,
			HardwareAcceleration.Mediatek,
			HardwareAcceleration.GoogleTensor,
		},
	},
	Platform.IOS: {
		Quantization.none: {
			# ?todo:
			HardwareAcceleration.CPU,
			HardwareAcceleration.XNNPACK,
			HardwareAcceleration.GPU,  # ?todo: == Metal
		},
		Quantization.int8: {
			HardwareAcceleration.CPU,
			HardwareAcceleration.XNNPACK,
			HardwareAcceleration.GPU,
		},
		Quantization.float16: {
			HardwareAcceleration.CPU,
			HardwareAcceleration.XNNPACK,
			HardwareAcceleration.GPU,
		},
	},
}

format = ModelFormat.LiteRT
format_distinguishes = [ModelAttributes.Quantization]



from pathlib import Path
from typing import Literal

from pydantic import BaseModel

from ai_parade._toolkit.data.ImageInput import ImageInput
from ai_parade._toolkit.data.output import Output
from ai_parade._toolkit.enums import ExactModelFormat


class ConversionResult(BaseModel):
	status: Literal["success", "failed"] = "success"

	format: ExactModelFormat
	"""
	Format of the converted model
	"""

	image_input: ImageInput
	"""
	Image input of the converted model
	"""

	path: Path
	"""
	Path to the converted model.
	"""
	output: Output
	"""
	Output of the converted model.
	"""

	@staticmethod
	def load(path: Path) -> "ConversionResult":
		if not path.exists():
			raise FileNotFoundError(f"Conversion result file {path} does not exist")
		if path.suffix != ".json":
			raise ValueError(f"Conversion result file {path} is not a JSON file")
		with open(path, "r", encoding="utf-8") as f:
			conversion_result = ConversionResult.model_validate_json(f.read())

		conversion_result.path = (path.parent / conversion_result.path).resolve()
		return conversion_result



import numpy as np
import ncnn
import torch

def test_inference():
    torch.manual_seed(0)
    in0 = torch.rand(1, 3, 8, 8, dtype=torch.float)
    out = []

    with ncnn.Net() as net:
        net.load_param("build/model.ncnn.param")
        net.load_model("build/model.ncnn.bin")

        with net.create_extractor() as ex:
            ex.input("in0", ncnn.Mat(in0.squeeze(0).numpy()).clone())

            _, out0 = ex.extract("out0")
            out.append(torch.from_numpy(np.array(out0)).unsqueeze(0))

    if len(out) == 1:
        return out[0]
    else:
        return tuple(out)

if __name__ == "__main__":
    print(test_inference())



import ai_parade._compat  # noqa: F401
import ai_parade._toolkit.runtime_metadata as runtime_metadata

from ._toolkit.data.ImageInput import ImageInput
from ._toolkit.data.output import ModelOutput
from ._toolkit.enums import AIRuntime, ExactModelFormat, ModelFormat
from ._toolkit.install._PackageManagerInstaller import PackageManagerInstaller
from ._toolkit.install._Uv import UV
from ._toolkit.install.ModelInstaller import ModelInstaller
from ._toolkit.metadata.enums import ModelTasks
from ._toolkit.metadata.models.final import (
	InstallOptionsApi,
	ModelMetadata,
	ModelMetadataApi,
	PyTorchOptionsApi,
)
from ._toolkit.metadata.size import parse_model_size
from ._toolkit.run.ModelRunner import ModelRunner
from ._toolkit.run.PyTorch import PyTorchRunner

__all__ = [
	# Installers
	"ModelInstaller",
	"PackageManagerInstaller",
	"UV",
	# Runners
	"ModelRunner",
	"PyTorchRunner",
	# Models
	"ModelOutput",
	"ModelTasks",
	"ModelFormat",
	"ExactModelFormat",
	"AIRuntime",
	"ImageInput",
	"ModelMetadata",
	"ModelMetadataApi",
	"InstallOptionsApi",
	"PyTorchOptionsApi",
	# Utils
	"parse_model_size",
	"runtime_metadata",
]



from typing import (
	Callable,
	Iterable,
	override,
)

import numpy as np

from ai_parade._toolkit.data.output import ModelOutput

from .Dataset import Dataset, DatasetInfo, Output_unbatched


class IteratorDataset(Dataset[Output_unbatched]):
	"""Dataset wrapper for an iterable"""

	def __init__(
		self, iterable: Callable[[], Iterable[tuple[np.ndarray, ModelOutput]]]
	):
		super().__init__()
		self._source_iterable = iterable

	@override
	def walk(self) -> Iterable[Output_unbatched]:
		for i, (data, annotations) in enumerate(self._source_iterable()):
			info = DatasetInfo(
				source_data_shape=data.shape,
				data_source=f"{type(self).__name__}[{i}]",
				data_shape=data.shape,
			)
			yield data, annotations, info

	@override
	def parse_data(
		self, iterable: Iterable[Output_unbatched], should_load: bool
	) -> Iterable[Output_unbatched]:
		return iterable



from dataclasses import dataclass
from enum import StrEnum
from typing import List, Optional


@dataclass(frozen=True)
class ImageInput:
	"""
	Represents the input specification for a CV model.
	"""

	batchSize: int
	"""Number of images in a single batch."""

	height: int
	"""Height of the input image."""

	width: int
	"""Width of the input image."""

	channelOrder: "ImageInput.ChannelOrder"
	"""Order of color channels in the image."""

	dataOrder: "ImageInput.DataOrder"
	"""Order of data dimensions."""

	dataType: "ImageInput.DataType"
	"""Data type of the image pixels."""

	means: Optional[tuple[float, float, float] | tuple[float, float, float, float]] = (
		None
	)
	"""
	Mean values for normalization, per channel.
	In [0, 255] range for integer data (`dataType=uint8`).
	In [0, 1] range for float data, (`dataType=float32` or `dataType=float16`).
	"""

	stds: Optional[tuple[float, float, float] | tuple[float, float, float, float]] = (
		None
	)
	"""
	Standard deviation values for normalization, per channel.
	In [0, 255] range for integer data (`dataType=uint8`).
	In [0, 1] range for float data, (`dataType=float32` or `dataType=float16`).
	"""

	@property
	def channels(self) -> int:
		"""
		Number of channels in the image.
		"""
		return (
			4
			if self.channelOrder
			in {ImageInput.ChannelOrder.RGBA, ImageInput.ChannelOrder.BGRA}
			else 3
		)

	@property
	def shape(self) -> List[int]:
		"""
		Shape of the input image, including batch size
		"""
		return (
			[self.batchSize, self.channels, self.height, self.width]
			if self.dataOrder == ImageInput.DataOrder.NCHW
			else [self.batchSize, self.height, self.width, self.channels]
		)

	class DataOrder(StrEnum):
		"""Enumeration for data order formats."""

		NHWC = "NHWC"
		NCHW = "NCHW"

	class DataType(StrEnum):
		"""Enumeration for data types."""

		uint8 = "uint8"
		float32 = "float32"
		float16 = "float16"

	class ChannelOrder(StrEnum):
		"""Enumeration for channel order types."""

		RGB = "RGB"
		BGR = "BGR"
		RGBA = "RGBA"
		BGRA = "BGRA"



from pathlib import Path

from ai_parade._toolkit.enums import AIRuntime, formats_preferred_runtime
from ai_parade._toolkit.metadata.custom import Customizations
from ai_parade._toolkit.metadata.models.api import RunnerGetter
from ai_parade._toolkit.metadata.models.final import ModelMetadata

from .ModelRunner import ModelRunner
from .NotARunner import NotARunner
from .RemoteRunner import RemoteRunner


def get_runner(
	model_metadata: ModelMetadata,
	weights: Path,
	runtime: AIRuntime | None = None,
	use_remote: bool = True,
	customizations: Customizations | None = None,
	runner_name: str | None = None,
) -> ModelRunner:
	"""Returns appropriate runner for the model based on the parameters

	Args:
		model_metadata: metadata
		weights: Model weights
		runtime: Desired AI runtime
		use_remote: If False, don't use remote runner. Defaults to True.
		customizations: Customizations. Defaults to None.
		runner_name: Name of the runner in the customizations. Defaults to None, required if customizations is not None

	Returns:
		Initialized ModelRunner
	"""
	runtime = runtime or formats_preferred_runtime[model_metadata.format.format]
	if model_metadata.get_custom_runner is not None:
		assert formats_preferred_runtime[model_metadata.format.format] == runtime
		runner = model_metadata.get_custom_runner(model_metadata, weights)
		if not use_remote and isinstance(runner, RemoteRunner):
			return runner.local_runner(model_metadata, weights)
		return runner
	else:
		local_getter = _choose_local_runner(runtime, customizations, runner_name)
		if not use_remote:
			return local_getter(model_metadata, weights)

		return RemoteRunner(model_metadata, runtime, weights, local_getter)


def _choose_local_runner(
	runtime: AIRuntime,
	customizations: Customizations | None = None,
	runner_name: str | None = None,
) -> RunnerGetter:
	"""Chooses default runner or runner from customizations

	Args:
		model_format: Desired model format
		customizations: Customizations. Defaults to None.
		runner_name: Name of the runner in the customizations. Defaults to None, required if customizations is not None

	Returns:
		ModelRunner in form of RunnerGetter
	"""
	if customizations is not None and runner_name is not None:
		return customizations.runners[runner_name]

	else:
		match runtime:
			case AIRuntime.PyTorch:
				from .PyTorch import PyTorchRunner

				return PyTorchRunner
			case AIRuntime.ONNXRuntime:
				from .ONNX import ONNXRunner

				return ONNXRunner
			case AIRuntime.TensorFlow:
				from .TensorFlow import TensorFlowRunner

				return TensorFlowRunner
			case AIRuntime.ExecuTorch:
				return NotARunner
			case AIRuntime.LiteRT:
				return NotARunner
			case AIRuntime.ncnn:
				return NotARunner
			case _:
				raise RuntimeError(f"Unknown model format `{runtime}`")
			# New AI Runtime: register runner - ModelRunner derived class



from pathlib import Path
from typing import TYPE_CHECKING, Callable

from ai_parade._toolkit.enums.formats import ExactModelFormat
from ai_parade._toolkit.install.ModelInstaller import ModelInstaller
from ai_parade._toolkit.run.ModelRunner import ModelRunner

from .pyproject import (
	InstallOptionsPyproject,
	ModelMetadataPyproject,
	PyTorchOptionsPyproject,
)

if TYPE_CHECKING:
	# its circular import, but we need it only for types and we could split the models
	from .final import ModelMetadata

InstallerGetter = Callable[["ModelMetadata"], ModelInstaller]
RunnerGetter = Callable[["ModelMetadata", Path], ModelRunner]


class ModelMetadataApi(ModelMetadataPyproject):
	"""
	Api definition of model metadata (registration).
	"""

	get_installer: InstallerGetter | None = None
	"""
	Uninitialized type of the installer
	None (default) will use default `CondaPipInstaller`
	"""

	get_custom_runner: RunnerGetter | None = None
	"""
	Uninitialized type of the runner

	The getter function which returns ModelRunner instance from ModelMetadata and weights path.

	None (default) will use default runner for given format
	"""

	size: float  # type: ignore redefinition with un-optional type

	format: ExactModelFormat  # type: ignore redefinition with exact format


InstallOptionsApi = InstallOptionsPyproject
"""
Api definition of install options 
"""
PyTorchOptionsApi = PyTorchOptionsPyproject
"""
Api definition of pytorch options 
"""



import logging
from typing import override

from ai_parade._toolkit.constants import TRACE_LVL, VENV_DIR
from ai_parade._toolkit.logging import subprocess_run_log_async
from ai_parade.toolkit import ModelMetadata

from ._PackageManagerInstaller import PackageManagerInstaller
from .ModelInstaller import ModelInstaller

logger = logging.getLogger(__name__)


class CondaInstaller(ModelInstaller):
	"""Installs dependencies with conda and pip

	The installation process is as follows:
	- download model (model weights)
	- download repository (git clone)
	- if conda is available:
		- create venv with conda, set up proper python version
		- install conda dependencies from ai-parade.install.conda_dependencies in pyproject.toml
	- if conda is not available:
		- create venv
	- install pip dependencies:
		- install dependencies from ai-parade.install.pip_dependencies in pyproject.toml
		- install dependencies from requirements.txt
		- install self, this will invoke build scripts like setup.py, pyproject.toml, etc.
	"""

	def __init__(
		self,
		model_metadata: ModelMetadata,
		package_installer: PackageManagerInstaller,
	):
		"""Interacts with a model

		Args:
				 model_metadata: metadata
				 model: Fully initialized model, including weights Default None.
		"""

		self.model_metadata = model_metadata
		self.conda = "conda"

		assert model_metadata.repository_path is not None
		self.repo_path = model_metadata.repository_path
		self.venv_path = model_metadata.repository_path / VENV_DIR

		package_installer.package_manager.prepend_commands(
			[self.conda, "run", "--prefix", str(self.venv_path)]
		)
		self.package_installer = package_installer

		assert model_metadata.install is not None
		self.install_parameters = model_metadata.install

		assert model_metadata.install.conda_dependencies is not None
		self.conda_dependencies = model_metadata.install.conda_dependencies

	async def is_conda_dependencies_installed(self) -> bool:
		"""Checks if conda dependencies are installed

		Returns:
			 True if conda dependencies are installed
		"""

		logger.log(TRACE_LVL, "Checking if conda dependencies are installed")

		retCode, stdout, _ = await subprocess_run_log_async(
			[self.conda, "install", "--dry-run", "--prefix", self.venv_path],
			check=False,
			logger=logger,
		)
		is_installed = stdout.find("All requested packages already installed.") != -1
		return retCode == 0 and is_installed

	@override
	async def is_installed(self) -> bool:
		"""Checks if model and dependencies are installed

		Returns:
			 True if everything is installed
		"""

		return (
			self.package_installer.is_installed()
			and await self.is_conda_dependencies_installed()
		)

	def is_venv_created(self) -> bool:
		"""Checks if venv could be found

		Returns:
			 True if venv is found
		"""

		logger.log(TRACE_LVL, "Checking if venv is created")
		return (self.venv_path / "conda-meta").is_dir()

	async def _create_conda_venv(self):
		"""Creates conda environment"""
		if self.is_venv_created():
			logger.debug("Conda environment already there. Skipping")
			return

		if (self.repo_path / "environment.yml").exists():
			logger.debug("Creating conda environment from environment.yml")
			await subprocess_run_log_async(
				[
					self.conda,
					"create",
					"--prefix",
					self.venv_path,
					"--file",
					"environment.yml",
					"--yes",
					"--quiet",
				],
				logger=logger,
			)
		else:
			logger.debug("Creating conda venv")
			await subprocess_run_log_async(
				[
					self.conda,
					"create",
					"--prefix",
					self.venv_path,
					"pip",
					"python=" + self.install_parameters.python,
					"--yes",
					"--quiet",
				],
				logger=logger,
			)

	async def _install_conda_dependencies(self):
		"""Installs conda dependencies from ai-parade.install.conda_dependencies"""
		if await self.is_conda_dependencies_installed():
			logger.debug("Conda dependencies already installed. Skipping")
			return

		logger.debug("Installing conda dependencies")
		await subprocess_run_log_async(
			[
				self.conda,
				"install",
				"--prefix",
				self.venv_path,
				"--yes",
				"--quiet",
			]
			+ list(self.conda_dependencies),
			logger=logger,
		)

	@override
	async def install(self, override: bool = False):
		"""Installs model see `CondaPipInstaller`

		Args:
			 override: If True, will override all existing files. Defaults to False.
		"""
		logger.info(f"Installing model `{self.model_metadata.name}`: ")

		logger.info("Going to install model dependencies")
		# todo: nuke the dependencies on override == True
		await self._create_conda_venv()
		await self._install_conda_dependencies()
		await self.package_installer._install_pip_dependencies()

		logger.info("Installation done.")



import functools
from typing import Callable, Literal, override

from ai_parade._toolkit.enums import ExactModelFormat, ModelFormat, Quantization
from ai_parade._toolkit.run import ModelRunner
from ai_parade._toolkit.run.ONNX import ONNXRunner

from .converter import ConversionResult, Converter


class ONNX_Quantization(Converter):
	"""
	Built-in ONNX post training quantization
	"""

	def __init__(
		self,
		quantization_type: Literal[Quantization.float16] | Literal[Quantization.int8],
	):
		self.quantization_type = quantization_type

	@property
	@override
	def conversion(self):
		return (
			ExactModelFormat(ModelFormat.ONNX),
			ExactModelFormat(ModelFormat.ONNX, self.quantization_type),
		)

	@override
	def create(
		self, options: Converter.Options
	) -> Callable[[ModelRunner], ConversionResult]:
		super().create(options)
		match self.quantization_type:
			case Quantization.float16:
				return self.convert_to_f16
			case Quantization.int8:
				assert options.calibration_data is not None
				return functools.partial(self.convert_to_i8, options=options)
			case _:
				raise RuntimeError("Unknown quantization type")

	def convert_to_f16(self, runner: ModelRunner):
		"""
		https://onnxruntime.ai/docs/performance/model-optimizations/float16.html
		"""
		assert isinstance(runner, ONNXRunner)
		import onnx  # type: ignore
		from onnxconverter_common import float16  # type: ignore

		# ?todo: https://onnxruntime.ai/docs/performance/transformers-optimization.html

		model = onnx.load(runner.weights)
		model_fp16 = float16.convert_float_to_float16(model, keep_io_types=True)
		output_model_path = self.get_output_weights_path(runner, None)
		onnx.save(model_fp16, output_model_path)

		return ConversionResult(
			format=self.conversion[1],
			image_input=runner.model_metadata.image_input,
			path=output_model_path,
			output=runner.model_metadata.output,
		)

	def convert_to_i8(self, runner: ModelRunner, options: Converter.Options):
		"""
		https://onnxruntime.ai/docs/performance/model-optimizations/quantization.html
		"""
		assert isinstance(runner, ONNXRunner)

		from onnxruntime.quantization import (  # type: ignore
			CalibrationDataReader,
			quant_pre_process,
			quantize_static,
		)

		class DataReader(CalibrationDataReader):
			def __init__(self, runner: ModelRunner):
				assert options.calibration_data is not None
				self.iterator = iter(
					map(lambda x: {"input": x[0]}, options.calibration_data)
				)

				# Use inference session to get input shape.
				# session = runner.model
				# (_, _, height, width) = session.get_inputs()[0].shape

				# self.input_name = session.get_inputs()[0].name

			def get_next(self):
				return next(self.iterator, {})

		# ?todo: https://onnxruntime.ai/docs/performance/transformers-optimization.html

		preprocessed_model = runner.output_directory / "preprocessed_model.onnx"
		quant_pre_process(runner.weights, preprocessed_model)
		quantize_static(
			preprocessed_model,
			self.get_output_weights_path(runner, options),
			calibration_data_reader=DataReader(runner),
		)

		return ConversionResult(
			format=self.conversion[1],
			image_input=runner.model_metadata.image_input,
			path=self.get_output_weights_path(runner, options),
			output=runner.model_metadata.output,
		)



from abc import ABC, abstractmethod
from typing import Any, override


class ModelInstaller(ABC):
	"""Interface of model installer"""

	def __init__(self, *args: Any) -> None:
		pass

	async def is_installed(self) -> bool:
		"""Checks if model and dependencies are installed

		Returns:
			True if everything is installed
		"""
		return False

	@abstractmethod
	async def install(self, override: bool = False) -> None:
		"""Installs model and all its dependencies

		Args:
			 override: if True, will override all existing files. Defaults to False.
		"""
		pass


class NOPInstaller(ModelInstaller):
	"""Does nothing"""

	@override
	async def install(self, override: bool = False) -> None:
		pass

	@override
	async def is_installed(self) -> bool:
		return True



from dataclasses import replace
from typing import TYPE_CHECKING, Literal, override

if TYPE_CHECKING:
	import torch  # type: ignore  # noqa: F401

from ai_parade._toolkit.data.ImageInput import ImageInput
from ai_parade._toolkit.enums import (
	ExactModelFormat,
	ModelFormat,
	Quantization,
)

from ._pytorch_base import PyTorchConvertBase
from .converter import Converter


class AIEdgeConverterEntry(Converter):
	"""
	https://github.com/google-ai-edge/ai-edge-torch
	and
	https://github.com/google-ai-edge/ai-edge-torch/blob/main/docs/pytorch_converter/README.md#quantization
	"""

	def __init__(
		self,
		quantization_type: Literal[Quantization.none]
		| Literal[Quantization.float16]
		| Literal[Quantization.int8],
	):
		self.quantization_type = quantization_type

	@property
	@override
	def conversion(self):
		return (
			ExactModelFormat(ModelFormat.PyTorch),
			ExactModelFormat(ModelFormat.LiteRT, self.quantization_type),
		)

	@override
	def create(self, options: Converter.Options):
		super().create(options)
		return self.AIEdgeConverter(self.conversion, options)

	class AIEdgeConverter(PyTorchConvertBase):
		@override
		def _export_call(self, params: PyTorchConvertBase.Params):
			model, sample_inputs, output_weights_path, default_results, *_ = params

			import ai_edge_torch  # type: ignore
			import tensorflow as tf  # type: ignore

			# set it to channels last mode, so that the model does not contain transpose layers (the backend does work in NHWC)
			model = ai_edge_torch.to_channel_last_io(model, args=[0])
			model.eval()  # apparently the mode is reset by the to_channel_last_io
			sample_inputs = (sample_inputs[0].permute(0, 2, 3, 1),)

			converter_flags = {}
			match self.conversion[1].quantization:
				case Quantization.float32:
					pass

				case Quantization.int8:
					assert self.options.calibration_data is not None
					# with fallback to other datatypes
					converter_flags["optimizations"] = [tf.lite.Optimize.DEFAULT]
					converter_flags["representative_dataset"] = list(
						map(
							lambda x: [x[0].transpose(0, 2, 3, 1)],
							self.options.calibration_data,
						)
					).__iter__
				case Quantization.float16:
					converter_flags = {
						"optimizations": [tf.lite.Optimize.DEFAULT],
						"target_spec": {"supported_types": [tf.float16]},
					}

				case _:
					raise NotImplementedError(
						f"Quantization {self.conversion[1].quantization} not supported"
					)

			edge_model = ai_edge_torch.convert(
				model, sample_inputs, _ai_edge_converter_flags=converter_flags
			)
			edge_model.export(str(output_weights_path))

			default_results.image_input = replace(
				default_results.image_input, dataOrder=ImageInput.DataOrder.NHWC
			)
			return default_results


capabilities: list[Converter] = [
	AIEdgeConverterEntry(Quantization.none),
	AIEdgeConverterEntry(Quantization.int8),
	AIEdgeConverterEntry(Quantization.float16),
]



from abc import ABC, abstractmethod
from pathlib import Path
from typing import TYPE_CHECKING, Any, Callable, Generic, ParamSpec, TypeVar

from ai_parade._toolkit.datasets.Dataset import Dataset

if TYPE_CHECKING:
	from ai_parade._toolkit.metadata.models.final import ModelMetadata
from ai_parade._toolkit.data.output import ModelOutput

TChild = TypeVar("TChild", bound="LoadedModelRunner")
TReturn = TypeVar("TReturn")
TModel = TypeVar("TModel")
P = ParamSpec("P")


# backward compatibility <=3.11: generic class notation
class ModelRunner(Generic[TChild, TModel], ABC):
	"""
	Runs model inference or anything on loaded model. It is a wrapper of execution runtimes.

	Implementation notes:
	- **IMPORTANT**: Do not import the runtime on module level, the module should be importable even if the runtime is not installed.
		e.g. do not do `import torch` at the top of the module
	"""

	def __init__(
		self, model_metadata: "ModelMetadata", weights: Path, *args: Any
	) -> None:
		self._output_dir: Path = Path.cwd()
		self._model_metadata = model_metadata
		self.weights = weights

	@property
	def output_directory(self) -> Path:
		"""Used for logs, artifacts, etc."""
		return self._output_dir

	@property
	def model_metadata(self) -> "ModelMetadata":
		return self._model_metadata

	async def get_model_size(self) -> int | None:
		"""
		Get the size of the model in bytes

		Returns:
			Size of the model in bytes or None if the size could not be determined (model not yet downloaded)
		"""
		if not self.weights.exists():
			return None
		return self.weights.resolve().stat().st_size

	async def set_output_directory(self, value: Path | None):
		"""Set output directory for the runner. Used for logs, artifacts, etc."""
		self._output_dir = value or Path.cwd()

	@abstractmethod
	async def __aenter__(self) -> "LoadedModelRunner[TChild, TModel]":
		pass

	@abstractmethod
	async def __aexit__(self, exc_type: Any, exc_value: Any, traceback: Any):
		pass


class LoadedModelRunner(Generic[TChild, TModel], ABC):
	"""
	Runs model inference or anything on loaded model
	"""

	def __init__(self, runner: ModelRunner) -> None:
		self._runner = runner

	@property
	def output_directory(self) -> Path:
		"""Used for logs, artifacts, etc."""
		return self._runner._output_dir

	@property
	def model_metadata(self) -> "ModelMetadata":
		return self._runner._model_metadata

	async def get_model_size(self) -> int | None:
		"""
		Get the size of the model in bytes

		Returns:
			Size of the model in bytes
		"""
		size = await self._runner.get_model_size()
		assert size is not None
		return size

	async def set_output_directory(self, value: Path | None):
		"""Set output directory for the runner. Used for logs, artifacts, etc."""
		# it is not property setter because I am lazy to implement them in the puppet
		await self._runner.set_output_directory(value)

	@abstractmethod
	async def inference(self, dataset: Dataset) -> list[ModelOutput]:
		"""Run inference on model

		Args:
			dataset: data to use to the inference

		Returns:
			output of the model
		"""
		pass

	@abstractmethod
	async def run(self, what: Callable[[TChild], TReturn]) -> TReturn:
		"""Run something in the scope of the runner

		For example the remote runner is given converter to convert the model

		Args:
			what: callable to run, it is given the model from the runner

		Returns:
			result of the callable
		"""
		pass

	@property
	@abstractmethod
	def model(self) -> TModel:
		"""Get internal representation of the model (i.e. PyTorch model)

		Used by the converters
		"""
		pass



from ..enums import (
	AIRuntime,
	HardwareAcceleration,
	ModelFormat,
	Platform,
	Quantization,
)
from ._common import ModelAttributes

runtime = AIRuntime.ncnn
capabilities = {
	Platform.Android: {
		Quantization.none: {
			HardwareAcceleration.CPU,
			HardwareAcceleration.Vulkan,
		},
		Quantization.int8: {
			HardwareAcceleration.CPU,
			HardwareAcceleration.Vulkan,
		},
		Quantization.float16: {
			HardwareAcceleration.CPU,
			HardwareAcceleration.Vulkan,
		},
	},
	Platform.IOS: {
		Quantization.none: {
			HardwareAcceleration.CPU,
			HardwareAcceleration.Vulkan,
		},
		Quantization.int8: {
			HardwareAcceleration.CPU,
			HardwareAcceleration.Vulkan,
		},
		Quantization.float16: {
			HardwareAcceleration.CPU,
			HardwareAcceleration.Vulkan,
		},
	},
}

format = ModelFormat.NCNN
format_distinguishes = [ModelAttributes.Quantization]



from typing import TYPE_CHECKING, Any, override

from ai_parade._toolkit.run.ONNX import ONNXRunner

if TYPE_CHECKING:
	import torch  # type: ignore  # noqa: F401

from ai_parade._toolkit.enums import ModelFormat

from ._onnx_simplify import ONNXSimplify
from ._pytorch_base import PyTorchConvertBase
from .converter import Converter, FormatConverter


class PyTorch_onnx_export(FormatConverter):
	"""
	Built-in converter from PyTorch to ONNX.
	"""

	@property
	@override
	def format_conversion(self):
		return (ModelFormat.PyTorch, ModelFormat.ONNX)

	@override
	def create(self, options: Converter.Options):
		super().create(options)
		return self.PyTorchConverter(self.conversion, options)

	class PyTorchConverter(PyTorchConvertBase):
		@override
		def _export_call(self, params: PyTorchConvertBase.Params):
			model, sample_inputs, output_weights_path, *_ = params

			import torch  # type: ignore
			from onnxscript import opset18 as op  # type: ignore

			def sym_not(x: Any):
				return op.Not(x)

			# This notation is kind of cryptic but:
			# - tuple define model input argument (could be dict instead to act as kwargs)
			# - key in the dictionary is the index of its dimensions
			# - value is the dimension
			dynamic_shapes = ({0: torch.export.Dim("batch_dim", min=1)},)

			onnx_model = torch.onnx.export(
				model,
				sample_inputs,
				external_data=False,  # ?todo: at least for now this large models are useless for us
				dynamo=True,
				report=True,
				artifacts_dir=output_weights_path.parent,
				verify=True,
				dynamic_shapes=dynamic_shapes,
				input_names=[ONNXRunner.INPUT_NAME],
				# todo after https://github.com/pytorch/pytorch/issues/136572 : remove
				custom_translation_table={torch.sym_not: sym_not},
			)
			if not onnx_model:
				raise RuntimeError("Failed to export ONNX model")
			onnx_model.save(output_weights_path)

			onnx_model.release()
			onnx_model = None

			if not ONNXSimplify().simplify(output_weights_path):
				raise RuntimeError("Failed to simplify ONNX model")



from ai_parade._toolkit.enums import ExactModelFormat


class ConversionError(RuntimeError):
	def __init__(
		self,
		source: ExactModelFormat,
		target: ExactModelFormat,
		message: str = "",
	):
		self.source = source
		self.target = target
		self.message = message
		super().__init__(self.__str__())

	def __str__(self) -> str:
		return f"Conversion from {self.source} to {self.target} failed:\n{self.message}"



import functools
import logging
import re
from typing import Any, Callable, cast

logger = logging.getLogger(__name__)

# ?todo: Recursive definition not working with pydantic
MappingDefinition0 = str | list[Any] | dict[str, Any]
MappingDefinition1 = str | list[MappingDefinition0] | dict[str, MappingDefinition0]
MappingDefinition = str | list[MappingDefinition1] | dict[str, MappingDefinition1]
# MappingDefinition = str | list["MappingDefinition"] | dict[str, "MappingDefinition"]


class StructuralMappingError(Exception):
	pass


def _get_repeat_pattern(mapping: list | tuple) -> tuple[int, str | None]:
	for idx, item in enumerate(mapping):
		if isinstance(item, str) and item.startswith("...repeat"):
			match = re.match(r"\.\.\.repeat\s+as\s+(?P<name>\w+)", item)
			if match:
				return idx, match.group("name")
			else:
				return idx, ""
	return len(mapping), None


def structural_map(
	raw_output: Any,
	mapping: MappingDefinition | None,
	skip_missing: bool = False,
) -> dict[str, float | int | str | list[float | int | str]]:
	if mapping is None:
		return cast(dict[str, Any], raw_output)

	def _recursive_map(
		raw_output: Any,
		mapping: Any,
		parsed_output: dict[str, Any] = {},
		can_repeat: bool = True,
	) -> dict[str, Any] | None:
		try:
			if isinstance(mapping, str):
				if mapping in ["...skip", ""]:
					# Skip this element
					return
				elif parsed_output.get(mapping) is None:
					parsed_output[mapping] = raw_output
				else:
					parsed_output[mapping] = (
						parsed_output[mapping]
						if isinstance(parsed_output[mapping], list)
						else [parsed_output[mapping]]
					)
					parsed_output[mapping].append(raw_output)
			elif isinstance(mapping, list) or isinstance(mapping, tuple):
				# Handle repeating patterns
				repeat_len, repeat_name = _get_repeat_pattern(mapping)
				if (
					repeat_len > len(raw_output)
					and "...repeat" not in mapping
					and not (len(raw_output) == 0 and repeat_name is not None)
				):
					raise StructuralMappingError(
						IndexError,
						raw_output,
						mapping,
						parsed_output,
					)
				if repeat_name is not None:
					if not can_repeat:
						raise StructuralMappingError(
							ValueError(
								"'...repeat' cannot be nested inside another repeating pattern."
							),
							raw_output,
							mapping,
							parsed_output,
						)
					if repeat_name == "":
						can_repeat = False
					else:
						parsed_output.setdefault(repeat_name, [])
				for raw_item_index, raw_item in enumerate(
					raw_output[: len(raw_output) // repeat_len * repeat_len]
				):
					mapping_index = raw_item_index % repeat_len
					repetition = raw_item_index // repeat_len

					if repeat_name is not None and repeat_name != "":
						if mapping_index == 0 and repetition >= len(
							parsed_output[repeat_name]
						):
							parsed_output[repeat_name].append({})

						_recursive_map(
							raw_item,
							mapping[mapping_index],
							parsed_output[repeat_name][repetition],
							can_repeat,
						)

					else:
						if repeat_name == "" and mapping[mapping_index] != "...skip":
							parsed_output.setdefault(mapping[mapping_index], [])
						_recursive_map(
							raw_item, mapping[mapping_index], parsed_output, can_repeat
						)

			else:
				if isinstance(raw_output, dict):
					for key in mapping.keys():
						_recursive_map(
							raw_output[key], mapping[key], parsed_output, can_repeat
						)
				else:
					for key in mapping.keys():
						_recursive_map(
							getattr(raw_output, key),
							mapping[key],
							parsed_output,
							can_repeat,
						)
		except (
			TypeError,  # wrong indexing
			IndexError,  # list indexing
			KeyError,  # dict indexing
			AttributeError,  # object attribute indexing
		) as e:
			if skip_missing:
				logger.debug(f"Skipping missing index `{e.args[0]}`")
			else:
				raise StructuralMappingError(e, raw_output, mapping, parsed_output)
		return parsed_output

	parsed_output = _recursive_map(raw_output, mapping) or {}

	return parsed_output


ObjectMapping = Callable[[Any], dict[str, Any] | list[dict[str, Any]]]


def structural_map_fx(
	mapping: MappingDefinition | None,
	skip_missing: bool = False,
) -> ObjectMapping:
	return functools.partial(
		structural_map,
		mapping=mapping,
		skip_missing=skip_missing,
	)



import shutil
from pathlib import Path
from typing import TYPE_CHECKING, Callable, Literal, override

from ai_parade._toolkit.converters._pytorch_base import PyTorchConvertBase
from ai_parade._toolkit.enums import ExactModelFormat, ModelFormat, Quantization
from ai_parade._toolkit.run import ModelRunner
from ai_parade._toolkit.run.ONNX import ONNXRunner

from .converter import ConversionResult, Converter

if TYPE_CHECKING:
	import torch  # type: ignore  # noqa: F401


def _create_ncnn_model_dir(output_weights_path: Path):
	output_weights_path.mkdir(exist_ok=True)
	(output_weights_path.parent / "model.ncnn.bin").rename(
		output_weights_path / "model.bin"
	)
	(output_weights_path.parent / "model.ncnn.param").rename(
		output_weights_path / "model.param"
	)


class PNNXPytorchConverter(Converter):
	"""
	https://github.com/Tencent/ncnn/tree/master/tools/pnnx/python
	https://github.com/pnnx/pnnx
	"""

	def __init__(
		self, quantization: Literal[Quantization.none] | Literal[Quantization.float16]
	):
		self.quantization = quantization

	@property
	@override
	def conversion(self):
		return (
			ExactModelFormat(ModelFormat.PyTorch),
			ExactModelFormat(ModelFormat.NCNN, self.quantization),
		)

	@override
	def create(self, options: Converter.Options):
		super().create(options)
		return self.PNNXConverter(self.conversion, options)

	class PNNXConverter(PyTorchConvertBase):
		@override
		def _export_call(self, params: PyTorchConvertBase.Params):
			model, sample_inputs, output_weights_path, *_ = params

			import pnnx  # type: ignore

			_ = pnnx.export(
				model,
				f"{output_weights_path.parent}/model.pt",
				sample_inputs,
				fp16=self.conversion[1].quantization == Quantization.float16,
			)
			_create_ncnn_model_dir(output_weights_path)


class PNNXConverter(Converter):
	def __init__(self, conversion: tuple[ExactModelFormat, ExactModelFormat]):
		self._conversion = conversion

	@property
	@override
	def conversion(self):
		return self._conversion

	@override
	def create(
		self, options: Converter.Options
	) -> Callable[[ModelRunner], ConversionResult]:
		super().create(options)
		return self.convert

	def convert(self, runner: ModelRunner):
		assert isinstance(runner, ONNXRunner)
		import pnnx  # type: ignore
		import torch  # type: ignore

		shutil.copy(runner.weights, runner.output_directory / "model.onnx")
		# I guess it does not have an option to set output directory (for everything)
		_ = pnnx.convert(
			str(runner.output_directory / "model.onnx"),
			torch.rand(runner.model_metadata.image_input.shape),
			fp16=self.conversion[1].quantization != Quantization.float32,
		)
		_create_ncnn_model_dir(self.get_output_weights_path(runner, None))

		return ConversionResult(
			format=self.conversion[1],
			image_input=runner.model_metadata.image_input,
			path=self.get_output_weights_path(runner, None),
			output=runner.model_metadata.output,
		)


capabilities: list[Converter] = [
	PNNXPytorchConverter(Quantization.none),
	PNNXPytorchConverter(Quantization.float16),
	PNNXConverter(
		(
			ExactModelFormat(ModelFormat.ONNX),
			ExactModelFormat(ModelFormat.NCNN),
		)
	),
	PNNXConverter(
		(
			ExactModelFormat(ModelFormat.ONNX),
			ExactModelFormat(ModelFormat.NCNN, Quantization.float16),
		)
	),
]



import functools
from pathlib import Path
from typing import Any, cast

from ai_parade._toolkit.data.output import ModelOutput
from ai_parade._toolkit.structural_mapping import MappingDefinition, structural_map

from ._PascalVOC import PascalVOC
from .Dataset import DatasetSplit

id_to_info: dict[str, tuple[int, str]] | None = None
index_to_info: dict[int, tuple[str, str]] | None = None


def _map_ImageNet_to_ModelOutput(
	raw: Any, mapping: MappingDefinition, mapping_path: Path
) -> ModelOutput:
	if raw == {}:
		return {}
	if not id_to_info:
		_load_imageNet_mapping(mapping_path)
		assert id_to_info

	# make sure the detections are a list
	if raw.get("annotation", {}).get("object", None) is not None:
		if not isinstance(raw["annotation"]["object"], list):
			raw["annotation"]["object"] = [raw["annotation"]["object"]]
	else:
		raw["annotation"]["object"] = []

	mapped = cast(ModelOutput, structural_map(raw, mapping, skip_missing=True))
	if (
		"detections" in mapped
		and len(mapped["detections"]) == 0
		and "class_ImageNet1k_id" not in mapped
	):
		return mapped

	# use class id from detection if not present
	if (
		"class_ImageNet1k_id" not in mapped
		and "detections" in mapped
		and "class_ImageNet1k_id" in mapped["detections"][0]
	):
		mapped["class_ImageNet1k_id"] = mapped["detections"][0]["class_ImageNet1k_id"]

	# map class id to index: in both classification and detection
	if "class_ImageNet1k_id" in mapped:
		if mapped["class_ImageNet1k_id"] in id_to_info:
			mapped["class_ImageNet1k"] = id_to_info[mapped["class_ImageNet1k_id"]][0]
	if "detections" in mapped:
		for detection in mapped["detections"]:
			if "class_ImageNet1k_id" in detection:
				if detection["class_ImageNet1k_id"] in id_to_info:
					detection["class_ImageNet1k"] = id_to_info[
						detection["class_ImageNet1k_id"]
					][0]
	return mapped


def _load_imageNet_mapping(path: Path):
	global id_to_info, index_to_info
	id_to_info = {}
	index_to_info = {}

	with open(path, "r") as f:
		for index, line in enumerate(f):
			line = line.strip()
			if not line:
				continue

			wnid, labels = line.split(" ", 1)

			id_to_info[wnid] = (index, labels)
			index_to_info[index] = (wnid, labels)


def ImageNet_VID(
	path: Path = Path("datasets/ImageNet/ILSVRC2015_VID"),
	split: DatasetSplit = DatasetSplit.train,
):
	if (path / "ImageNet").exists():
		path = path / "ImageNet"
	if (path / "ILSVRC2015_VID").exists():
		path = path / "ILSVRC2015_VID"
	return PascalVOC(
		path,
		split,
		functools.partial(
			_map_ImageNet_to_ModelOutput,
			mapping={
				"annotation": {
					"object": [
						{
							"name": "class_ImageNet1k_id",
							"bndbox": {
								"xmin": "bbox_xmin_absolute",
								"xmax": "bbox_xmax_absolute",
								"ymin": "bbox_ymin_absolute",
								"ymax": "bbox_ymax_absolute",
							},
						},
						"...repeat as detections",
					]
				}
			},
			mapping_path=path.parent / "LOC_synset_mapping.txt",
		),
	)


def ImageNet(
	path: Path = Path("datasets/ImageNet/ILSVRC"),
	split: DatasetSplit = DatasetSplit.train,
):
	if (path / "ImageNet").exists():
		path = path / "ImageNet"
	if (path / "ILSVRC").exists():
		path = path / "ILSVRC"

	return PascalVOC(
		path,
		split,
		functools.partial(
			_map_ImageNet_to_ModelOutput,
			mapping={
				"annotation": {
					"object": [
						{
							"name": "class_ImageNet1k_id",
							"bndbox": {
								"xmin": "bbox_xmin_absolute",
								"xmax": "bbox_xmax_absolute",
								"ymin": "bbox_ymin_absolute",
								"ymax": "bbox_ymax_absolute",
							},
						},
						"...repeat as detections",
					]
				}
			},
			mapping_path=path.parent / "LOC_synset_mapping.txt",
		),
	)



from ..enums import (
	AIRuntime,
	HardwareAcceleration,
	ModelFormat,
	Platform,
	Quantization,
)
from ._common import ModelAttributes

runtime = AIRuntime.PyTorch
capabilities = {
	Platform.Windows: {
		Quantization.none: {
			HardwareAcceleration.CPU,
		},
	}
}

format = ModelFormat.PyTorch
format_distinguishes = [ModelAttributes.Quantization]



from pathlib import Path
from typing import override

from .FileDataset import FileDataset


class COCO(FileDataset):
	def __init__(
		self,
		data_dir: Path,
		annotations_path: Path,
	):
		self._data_dir = data_dir
		self._annotations_path = annotations_path

	@override
	def walk(self):
		"""Walk the dataset directory tree.

		Returns:
			 iterator of tuples of (path to the annotation file and path to the data file)
		"""
		raise NotImplementedError()



from typing import Iterable, override

import numpy as np

from ai_parade._toolkit.data.ImageInput import ImageInput

from .Dataset import Dataset, DatasetInfo, Output_unbatched


class Random(Dataset[None]):
	"""Dataset with random data of the ImageInput specifications"""

	def __init__(
		self,
		image_input: ImageInput | None = None,
		*,
		seed: int | None = None,
		size: int = 1000,
	):
		super().__init__()
		self._image_input = image_input
		if image_input is not None:
			self._batch_size = image_input.batchSize
		self._seed = seed
		self._size = size

	@override
	def walk(self) -> Iterable[None]:
		for _ in range(self._size):
			yield None

	@override
	def parse_data(
		self, iterable: Iterable[None], should_load: bool
	) -> Iterable[Output_unbatched]:
		assert self._image_input is not None, (
			"The Random dataset was constructed without an ImageInput or preprocess call"
		)
		rng = np.random.RandomState(self._seed)
		shape = self._image_input.shape[1:]
		for _ in iterable:
			match self._image_input.dataType:
				case ImageInput.DataType.float32:
					data = rng.rand(*shape).astype(np.float32)
				case ImageInput.DataType.uint8:
					data = rng.randint(
						0,
						255,
						size=shape,
						dtype=np.uint8,
					)
			yield (
				data,
				{},
				DatasetInfo(
					source_data_shape=shape,
					data_source="Random",
					data_shape=shape,
				),
			)

	@override
	def preprocess(self, image_input: ImageInput):
		self._image_input = image_input
		self._batch_size = image_input.batchSize
		return self



import importlib
import importlib.util
import logging
import sys
from contextlib import asynccontextmanager
from pathlib import Path
from types import ModuleType
from typing import TYPE_CHECKING, Any, Callable, final, override
from xmlrpc.client import boolean

from ai_parade._toolkit.metadata.models.pyproject import InlineCode
from ai_parade._toolkit.sys_path_checkpoint import SysPathCheckpoint

from .ModelRunnerBase import (
	Dataset,
	ModelMetadata,
	ModelRunnerBase,
	structural_map,
)

if TYPE_CHECKING:
	import torch  # type: ignore

logger = logging.getLogger(__name__)


class PyTorchRunner(ModelRunnerBase["PyTorchRunner", "torch.nn.Module"]):
	def __init__(self, model_metadata: ModelMetadata, weights: Path):
		"""Interacts with a model

		There are these methods for overriding:
		- `inference_call`: the call to the model
		- `postprocess`: the postprocessing of the output
		- `load_weights`: the loading of the weights
		- `get_yacs_config`: the loading of the yacs config

		Args:
			model_metadata: metadata
			weights: Path to the model weights
		"""

		super().__init__(model_metadata, weights)
		if model_metadata.pytorch is None:
			raise RuntimeError("No way of constructing model, no `pytorch` key")
		self.run_parameters = model_metadata.pytorch
		self._model = None

	@final
	@property
	@override
	def model(self):
		assert self._model is not None
		return self._model

	@final
	def _to_cpu_numpy(self, x: Any, to_cpu: bool = True, to_numpy: bool = True):
		import torch  # type: ignore

		if isinstance(x, torch.Tensor):
			x = x.detach()
			if to_cpu:
				x = x.to("cpu")
			if to_numpy:
				x = x.numpy()
			return x
		elif isinstance(x, dict):
			return {k: self._to_cpu_numpy(v) for k, v in x.items()}
		elif isinstance(x, list) or isinstance(x, tuple):
			return [self._to_cpu_numpy(v) for v in x]
		else:
			raise TypeError(f"Unsupported type: {type(x)}")

	def _inference_call(self, tensor: "torch.Tensor") -> "torch.Tensor":
		assert self.model is not None
		return self.model(tensor)

	@final
	@override
	async def inference(self, dataset: Dataset):
		logger.info("Running inference on model")
		import torch  # type: ignore

		# self.model.to("cuda")
		self.model.eval()
		outputs = []
		infos = []
		for x, _, data_info in dataset:
			tensor = torch.from_numpy(x)  # .to("cuda")
			with torch.no_grad():
				raw_output = self._inference_call(tensor)
			outputs.append(self._to_cpu_numpy(raw_output))
			infos.append(data_info)

		return self._postprocess(outputs, infos, self.model_metadata)

	@override
	async def get_model_size(self) -> int | None:
		downloaded_size = await super().get_model_size()
		if downloaded_size is not None:
			return downloaded_size
		async with self.not_downloaded():
			param_size = 0
			for param in self.model.parameters():
				param_size += param.nelement() * param.element_size()

			buffer_size = 0
			for buffer in self.model.buffers():
				buffer_size += buffer.nelement() * buffer.element_size()

			return param_size + buffer_size

	@asynccontextmanager
	async def not_downloaded(self):
		try:
			yield self.__parametrized_enter(should_load_weights=False)
		finally:
			await self.__aexit__(*sys.exc_info())

	@final
	@override
	async def __aenter__(self):
		"""
		Initialize the model. Load the weights.
		"""
		return self.__parametrized_enter(should_load_weights=True)

	def __parametrized_enter(self, should_load_weights: boolean):
		logger.debug(f"Loading PyTorch model with weights at {self.weights}")
		assert self.model_metadata.repository_path is not None

		self.sys_path_checkpoint = SysPathCheckpoint(
			self.model_metadata.repository_path
		)
		self.sys_path_checkpoint.__enter__()

		args: list[Any] = []
		if self.run_parameters.yacs_configs is not None:
			args.append(self._get_yacs_config(self.run_parameters.yacs_configs))

		if self.run_parameters.model_init_expression is not None:
			logger.debug(
				f"Running `model_init_expression: {self.run_parameters.model_init_expression}`"
			)
			code = self.run_parameters.model_init_expression
			# has to return the model, without weight initialization
			self._model = self.__load_weights(
				lambda module: eval(code), should_load_weights
			)
		elif self.run_parameters.model_init_fx is not None:
			logger.debug(
				f"Running `model_init_fx: {self.run_parameters.model_init_fx}`"
			)
			model_init_fx = self.run_parameters.model_init_fx
			self._model = self.__load_weights(
				lambda module: getattr(
					module or sys.modules[__name__],
					model_init_fx,
				)(*args),
				should_load_weights,
			)

		else:
			assert False, "It should fail the validation during parsing"

		return self

	@final
	@override
	async def __aexit__(self, *err: Any):
		self._model = None
		assert self.model_metadata.repository_path is not None
		self.sys_path_checkpoint.__exit__(*err)

	def __load_weights(
		self,
		init_fx: Callable[[ModuleType | None], "torch.nn.Module"],
		should_load_weights: boolean,
	) -> "torch.nn.Module":
		import torch  # type: ignore

		logger.debug("Loading the model weights")
		if self.run_parameters.model_import is not None:
			logger.debug(
				f"Importing `model_import: {self.run_parameters.model_import}`"
			)
			try:
				model_module = importlib.import_module(self.run_parameters.model_import)
			except ModuleNotFoundError as e:
				if self.run_parameters.model_import in str(e):
					logger.warning(
						f"Can't import as module\n  defined as metadata.model_import = {self.run_parameters.model_import}"
					)
					logger.info(f"Failed with error: {str(e)}")
					# todo: this will probably never fix anything, because there should be implicit namespace packages

					assert self.model_metadata.repository_path is not None
					import_components = self.run_parameters.model_import.split(".")
					spec = importlib.util.spec_from_file_location(
						import_components[-1],
						Path(
							self.model_metadata.repository_path,
							*import_components[:-1],
							import_components[-1] + ".py",
						),
					)
					if spec is not None:
						logger.debug(
							"Importing it as a file, the error could be caused by missing `__init__.py`"
						)
						model_module = importlib.util.module_from_spec(spec)
						spec.loader.exec_module(model_module)  # type: ignore
					else:
						raise e
				else:
					raise e

		else:
			model_module = None

		model = init_fx(model_module)
		if not should_load_weights:
			return model
		if not self.weights.exists():
			raise FileNotFoundError(
				f"Model weights not found: {self.weights.resolve()}"
			)
		loaded_data = torch.load(
			self.weights,
			weights_only=True,
			map_location="cpu",
		)
		state_dict = structural_map(
			loaded_data,
			self.run_parameters.load_mapping
			if self.run_parameters.load_mapping is not None
			else "state_dict",
		)["state_dict"]

		model.load_state_dict(state_dict)
		return model

	def _get_yacs_config(self, yacs_configs: tuple[str | InlineCode, ...]) -> Any:
		logger.debug("Loading the yacs configs")
		# todo: where is it checked?
		assert self.model_metadata.repository_path is not None

		base_config = yacs_configs[0]
		if not isinstance(base_config, str):
			raise RuntimeError()

		cfg = importlib.import_module(base_config).cfg  # todo: or not?
		for config_file in yacs_configs[1:]:
			if isinstance(config_file, str):
				cfg.merge_from_file(self.model_metadata.repository_path / config_file)
			else:
				exec(config_file["code"])
		return cfg



from enum import StrEnum


class HardwareAcceleration(StrEnum):
	"""Enumeration for hardware acceleration options."""

	CPU = "CPU"
	GPU = "GPU"
	XNNPACK = "XNNPACK"
	NNAPI = "NNAPI"
	QNN = "QNN"
	Vulkan = "Vulkan"
	OpenGLES = "OpenGLES"
	OpenCL = "OpenCL"
	Mediatek = "Mediatek"
	Metal = "Metal"
	CoreML = "CoreML"
	GoogleTensor = "GoogleTensor"
	Exynos = "Exynos"


COMMON_HW_ACCELERATIONS = {
	HardwareAcceleration.CPU,
	HardwareAcceleration.GPU,
	HardwareAcceleration.XNNPACK,
	HardwareAcceleration.NNAPI,
	HardwareAcceleration.Vulkan,
	HardwareAcceleration.OpenGLES,
	HardwareAcceleration.OpenCL,
	# Apple specific:
	HardwareAcceleration.Metal,
	HardwareAcceleration.CoreML,
}



from dataclasses import replace

from ai_parade._toolkit.enums import ExactModelFormat, ModelFormat, Quantization

from .ai_edge_torch import capabilities as ai_edge_torch_capabilities
from .converter import Converter
from .executorch import capabilities as executorch_capabilities
from .ncnn import ncnnConverter
from .nobuco import NobucoConverter
from .onnx import ONNX_Quantization
from .onnx2tf import ONNX2TensorFlowConverter
from .onnx_tf import ONNX_TensorFlowConverter
from .pnnx import capabilities as pnnx_capabilities
from .pytorch import PyTorch_onnx_export

_priority_list: list[Converter] = (
	[
		# top priority first
		# to quantized, to edge formats:
		ONNX_Quantization(Quantization.int8),
		ONNX_Quantization(Quantization.float16),
		PyTorch_onnx_export(),
		ncnnConverter(),
	]
	+ executorch_capabilities
	+ pnnx_capabilities
	+ ai_edge_torch_capabilities
	# 3rd party:
	+ [
		NobucoConverter(),
		ONNX2TensorFlowConverter(),
	]
	# not maintained:
	+ [
		ONNX_TensorFlowConverter(),
	]
)

# adjacency list, keyed by the source node/format
_converter_from: dict[ExactModelFormat, list[Converter]] = {
	ExactModelFormat(source_format, source_quantization): []
	for source_format in ModelFormat
	for source_quantization in Quantization
}
for converter in _priority_list:
	source_format, target_format = converter.conversion
	assert source_format is not None and target_format is not None
	_converter_from[source_format].append(converter)


def get_converter(
	source_format: ExactModelFormat, target_format: ExactModelFormat
) -> list[list[Converter]]:
	"""
	Gets the list of converters that can convert from `source_format` to `target_format`.
	Returns a list of conversion chains.
	A conversion chain is a list of conversions from `source_format` via intermediate formats to `target_format`.
	The chains are sorted from the most reasonable one (most likely to work and be correct), others are there as fallback.

	Args:
		source_format (ExactModelFormat): the source format
		target_format (ExactModelFormat): the target format

	Returns:
		list[list[Converter]]: the list of conversion chains
	"""
	source_formats = {
		source_format,
		replace(source_format, hw_acceleration_restriction=None),
		replace(source_format, chip_compatibility=None),
		replace(
			source_format,
			hw_acceleration_restriction=None,
			chip_compatibility=None,
		),
	}
	open_conversion_chains: list[tuple[ExactModelFormat, list[Converter]]] = [
		(x, []) for x in source_formats
	]

	conversion_chains: list[list[Converter]] = []

	# BFS:
	while len(open_conversion_chains) > 0:
		chain_target_format, this_chain = open_conversion_chains.pop()
		for converter in _converter_from.get(chain_target_format, []):
			new_chain_target_format = converter.conversion[1]

			# we found the target
			if (
				new_chain_target_format == target_format
				or (
					new_chain_target_format
					== replace(target_format, hw_acceleration_restriction=None)
				)
				or (
					new_chain_target_format
					== replace(target_format, chip_compatibility=None)
				)
				or (
					new_chain_target_format
					== replace(
						target_format,
						hw_acceleration_restriction=None,
						chip_compatibility=None,
					)
				)
			):
				conversion_chains.append(this_chain + [converter])
				break

			# we are in a loop
			if new_chain_target_format in this_chain:
				break

			open_conversion_chains.append(
				(new_chain_target_format, this_chain + [converter])
			)

	return conversion_chains



from enum import StrEnum


class AIRuntime(StrEnum):
	TensorFlow = "TensorFlow"
	PyTorch = "PyTorch"
	ExecuTorch = "ExecuTorch"
	LiteRT = "LiteRT"
	ncnn = "ncnn"
	ONNXRuntime = "ONNXRuntime"
	# New AI Runtime: update here



import json
from pathlib import Path
from typing import Any, Callable, Iterable

import numpy as np

from ai_parade._toolkit.data.output import ModelOutput
from ai_parade._toolkit.datasets.Dataset import Dataset
from ai_parade._toolkit.metadata.models.final import ModelMetadata
from ai_parade._toolkit.run.runner_selector import get_runner


async def inference(
	weights_path: Path,
	model_metadata: ModelMetadata,
	dataset: Dataset,
	# common
	artifacts: Path,
	output: Path | None,
	metrics: dict[str, Callable[[Iterable[ModelOutput], Iterable[ModelOutput]], Any]],
	use_remote: bool,
):
	"""
	Run inference on a model using a selected dataset and print results.

	Args:
		weights_path: Path to weights or a conversion result JSON
		model_metadata: ModelMetadata
		dataset: Dataset without preprocessing (preprocessing done in this method)
		artifacts: Path
		output: Path path to json file to store results
		metrics: dict[str, Callable[[Iterable[ModelOutput], Iterable[ModelOutput]], Any]]
		use_remote: bool
	"""
	assert output is None or output.suffix == ".json"

	if weights_path.suffix == ".json":
		model_metadata, weights_path = model_metadata.with_conversion_results(
			weights_path
		)
	dataset = dataset.preprocess(model_metadata.image_input)

	async with get_runner(
		model_metadata, weights_path, use_remote=use_remote
	) as runner:
		await runner.set_output_directory(artifacts)
		raw_results = await runner.inference(dataset)

	gt = [label for _, label, _ in dataset.no_batch_iterator()]

	results = {}
	if output is not None:
		for metric, metric_fx in metrics.items():
			results[metric] = metric_fx(raw_results, gt)

		class NumpyEncoder(json.JSONEncoder):
			def default(self, o):
				if isinstance(o, np.ndarray):
					return o.tolist()
				elif isinstance(o, np.generic):
					return o.item()
				return super().default(o)

		with open(output, "w") as f:
			json.dump(results, f, cls=NumpyEncoder)

	return results



from typing import TYPE_CHECKING, override

if TYPE_CHECKING:
	import torch  # type: ignore  # noqa: F401

from ai_parade._toolkit.enums import ModelFormat

from ._pytorch_base import PyTorchConvertBase
from .converter import Converter, FormatConverter


class NobucoConverter(FormatConverter):
	"""
	https://github.com/AlexanderLutsenko/nobuco
	"""

	@property
	@override
	def format_conversion(self):
		return (ModelFormat.PyTorch, ModelFormat.SavedModel)

	@override
	def create(self, options: Converter.Options):
		super().create(options)
		return self.NobucoConverter(self.conversion, options)

	class NobucoConverter(PyTorchConvertBase):
		@override
		def _export_call(self, params: PyTorchConvertBase.Params):
			model, sample_inputs, output_weights_path, default_results, *_ = params

			import nobuco  # type: ignore  # noqa: F401
			from nobuco import (  # type: ignore  # noqa: F401
				ChannelOrder,
				ChannelOrderingStrategy,
			)
			from nobuco.layers.weight import WeightLayer  # type: ignore  # noqa: F401

			keras_model = nobuco.pytorch_to_keras(
				model,
				args=sample_inputs,
				kwargs=None,
				inputs_channel_order=ChannelOrder.TENSORFLOW,
				outputs_channel_order=ChannelOrder.TENSORFLOW,
			)
			keras_model.save(output_weights_path, save_format="tf")



import functools
import logging
from pathlib import Path
from tempfile import TemporaryDirectory
from typing import override

import numpy as np

from ai_parade._toolkit.datasets.Dataset import Dataset
from ai_parade._toolkit.enums.formats import (
	ExactModelFormat,
	ModelFormat,
	Quantization,
)
from ai_parade._toolkit.logging import subprocess_run_log
from ai_parade._toolkit.run.ModelRunner import ModelRunner

from .converter import ConversionResult, Converter

logger = logging.getLogger(__name__)


class ncnnConverter(Converter):
	def __init__(self) -> None:
		super().__init__()

	@property
	@override
	def conversion(self):
		return (
			ExactModelFormat(ModelFormat.NCNN),
			ExactModelFormat(ModelFormat.NCNN, Quantization.int8),
		)

	@override
	def create(self, options: Converter.Options):
		assert options.calibration_data is not None
		return functools.partial(
			self.convert, calibration_data=options.calibration_data
		)

	def convert(
		self, runner: ModelRunner, calibration_data: Dataset
	) -> ConversionResult:
		with TemporaryDirectory() as d:
			shenanigans_dir = Path(d)
			calibration_files = []
			for i, (data, annotation, info) in enumerate(
				calibration_data.no_batch_iterator()
			):
				filename = shenanigans_dir / f"calibration_data{i}.npy"
				calibration_files.append(str(filename.resolve()))
				with open(filename, "wb") as f:
					np.save(f, data)
			with open(shenanigans_dir / "filelist.txt", "w") as f:
				f.write("\n".join(calibration_files))

			subprocess_run_log(
				[
					"ncnn2table",
					str(runner.weights / "model.param"),
					str(runner.weights / "model.bin"),
					str(shenanigans_dir / "filelist.txt"),
					str(shenanigans_dir / "calibration.table"),
					f"shape=[{','.join(map(str, reversed(data.shape)))}]",  # wtf is the shape reversed here (it does not work otherwise)
					"thread=8",
					"method=kl",
					"type=1",
				],
				logger=logger,
			)
			output_model_path = self.get_output_weights_path(runner, None)
			print(output_model_path)
			output_model_path.mkdir(parents=True, exist_ok=True)
			subprocess_run_log(
				[
					"ncnn2int8",
					str(runner.weights / "model.param"),
					str(runner.weights / "model.bin"),
					str(output_model_path / "model.param"),
					str(output_model_path / "model.bin"),
					str(shenanigans_dir / "calibration.table"),
				],
				logger=logger,
			)

		return ConversionResult(
			format=self.conversion[1],
			image_input=runner.model_metadata.image_input,
			path=output_model_path,
			output=runner.model_metadata.output,
		)



from abc import ABC, abstractmethod
from pathlib import Path
from typing import Callable, NamedTuple, final, override

from ai_parade._toolkit.datasets.Dataset import Dataset
from ai_parade._toolkit.enums import (
	ExactModelFormat,
	ModelFormat,
)
from ai_parade._toolkit.get_weights import get_weights_name
from ai_parade._toolkit.run.ModelRunner import ModelRunner

from .ConversionResult import ConversionResult

Conversion = tuple[ExactModelFormat, ExactModelFormat]
"""
A conversion is a tuple of (input_format, output_format)
"""


class Converter(ABC):
	"""
	Converts a model from one format to another (including quantization).

	"""

	class Options(NamedTuple):
		"""
		Options for a converter.
		"""

		calibration_data: Dataset | None = None
		"""
		Dataset to use for quantization calibration, unused if no quantization
		"""

	@property
	@abstractmethod
	def conversion(self) -> Conversion:
		"""
		Available conversions in form of (input_format, output_format)
		"""
		pass

	@abstractmethod
	def create(
		self, options: Options
	) -> Callable[[ModelRunner], ConversionResult | list[ConversionResult]]:
		"""
		Creates a converter callable.

		The callable takes a ModelRunner as input and does not return anything.
		The converted model is stored in ModelRunner.output_directory as a `model.suffix` file(s).

		Implementation notes:
		- Use `get_output_weights_path` to get the path for the converted model

		Args:
			 calibration_data: dataset to use for quantization calibration, unused if no quantization

		Returns:
			 Callable that takes a ModelRunner as input and converts the model with it.
			 It returns the path to the converted model inside the runner's output directory
		"""

	def get_output_weights_path(
		self, runner: ModelRunner, options: Options | None
	) -> Path:
		return runner.output_directory / get_weights_name(self.conversion[1])


class FormatConverter(Converter):
	"""
	Subclass of Converter that converts model formats without quantization.
	"""

	@property
	@final
	@override
	def conversion(self) -> Conversion:
		formats = self.format_conversion
		return (
			ExactModelFormat(formats[0]),
			ExactModelFormat(formats[1]),
		)

	@property
	@abstractmethod
	def format_conversion(self) -> tuple[ModelFormat, ModelFormat]:
		pass



import logging
from pathlib import Path

TRACE_LVL = 5
logging.addLevelName(TRACE_LVL, "TRACE")

VENV_DIR = ".venv"
CUSTOM_DEFINITIONS_FILE_NAME = "ai-parade.py"


def get_site_packages_path(repository_path: Path) -> Path:
	return (repository_path / VENV_DIR / "lib").glob(
		"python*"
	).__next__() / "site-packages"



from .AIRuntime import AIRuntime
from .formats import (
	ExactModelFormat,
	ModelFormat,
	Quantization,
	format_suffixes,
	formats_preferred_runtime,
)
from .hw_acceleration import COMMON_HW_ACCELERATIONS, HardwareAcceleration
from .platform import Platform

__all__ = [
	"ModelFormat",
	"Quantization",
	"ExactModelFormat",
	"format_suffixes",
	"formats_preferred_runtime",
	"HardwareAcceleration",
	"COMMON_HW_ACCELERATIONS",
	"AIRuntime",
	"Platform",
]



import asyncio
import logging
import sys
from pathlib import Path
from pprint import pprint
from typing import Annotated

import rich
import rich.traceback
import typer

import ai_parade._compat  # noqa: F401
from ai_parade._toolkit.install.utilities import check_missing_dependencies
from ai_parade.datasets import DatasetSplit, ImageNet, ImageNet_VID, Random
from ai_parade.logging import TRACE_LVL
from ai_parade.toolkit import (
	Convert,
	Dataset,
	ExactModelFormat,
	HardwareAcceleration,
	ModelFormat,
	Quantization,
	RemoteError,
	get_local_metadata,
	get_runner,
)
from ai_parade.toolkit import (
	get_ai_parade_logger as get_toolkit_logger,
)
from ai_parade.toolkit import inference as run_inference

logger = logging.getLogger(__name__)

app = typer.Typer()

RepositoryPathType = Annotated[
	Path,
	typer.Option(
		"--repository",
		help="Path to repository with pyproject.toml",
		file_okay=False,
		exists=True,
		resolve_path=True,
	),
]
WeightsPathTypeArgDef = typer.Argument(
	help="Path to the model file (all in one file), or directory (if multiple files are needed), or json conversion result.",
	exists=True,
	resolve_path=True,
)
WeightsPathTypeOptional = Annotated[Path | None, WeightsPathTypeArgDef]
WeightsPathType = Annotated[Path, WeightsPathTypeArgDef]

ModelNameTypeArgDef = typer.Argument(
	help="Name of model (as in the [tool.ai-parade.name] property in pyproject.toml file)"
)
ModelNameType = Annotated[str, ModelNameTypeArgDef]
ModelNameTypeOptional = Annotated[str | None, ModelNameTypeArgDef]

ArtifactsDirType = Annotated[
	Path,
	typer.Option(
		help="Directory for artifacts (where logs, dumps etc. will be stored).",
		file_okay=False,
	),
]

OutputTypeOptionDef = typer.Option(
	"--output", "-o", help="Output directory for converted weights."
)

OutputType = Annotated[Path, OutputTypeOptionDef]
OutputTypeOptional = Annotated[Path | None, OutputTypeOptionDef]

SeparateEnvironmentType = Annotated[
	bool,
	typer.Option(
		"--separate-environment",
		"--separate-env",
		"--separate-venv",
		help="Separates environment (python interpreter and venv) from the local one. Setting this will complicate debugging.",
		show_default="False - local environment",
	),
]

NEED_CALIBRATION = [Quantization.int4, Quantization.int8]

verbosity_level = 0


def _set_verbosity(verbose: "VerboseType"):
	toolkit_logger = get_toolkit_logger()
	this_logger = logging.getLogger("ai_parade._cli.ai_parade")
	global verbosity_level
	verbosity_level = verbose.count(True)

	if verbosity_level < 2:
		logging.basicConfig(
			format="[%(levelname)-9s%(asctime)s] %(message)s",
			datefmt="%H:%M:%S",
		)
	else:
		logging.basicConfig(
			format="[%(levelname)-9s%(asctime)s] %(filename)s#%(lineno)d - %(message)s",
			datefmt="%H:%M:%S",
		)
	try:
		match verbosity_level:
			case 0:
				this_logger.setLevel(logging.INFO)
			case 1:
				toolkit_logger.setLevel(logging.INFO)
				this_logger.setLevel(logging.DEBUG)
			case 2:
				toolkit_logger.setLevel(logging.DEBUG)
				this_logger.setLevel(TRACE_LVL)
			case 3:
				toolkit_logger.setLevel(TRACE_LVL)
				this_logger.setLevel(TRACE_LVL)
			case _:
				logging.getLogger().setLevel(TRACE_LVL)
	except Exception as e:
		logging.basicConfig(level=logging.INFO)
		this_logger.error(f"Failed to set verbosity level: {e}")


VerboseType = Annotated[
	list[bool],
	typer.Option(
		"--verbose",
		"-v",
		help="Enable verbose logging, repeat for more verbosity (current max: `-vvvv`)",
		callback=_set_verbosity,
	),
]


def handle_remote_error(e: RemoteError):
	rich.print(e.rich_traceback)
	if verbosity_level >= 3:
		rich.print(rich.traceback.Traceback.from_exception(*sys.exc_info()))  # type: ignore
	raise typer.Exit(code=1)


def _get_dataset(
	dataset_name: str, dataset_path: Path | None, split: DatasetSplit
) -> Dataset:
	match dataset_name:
		case "ImageNet":
			if dataset_path is None:
				raise ValueError("Dataset path is required for ImageNet dataset")
			dataset = ImageNet(dataset_path, split)
		case "ImageNet_VID":
			if dataset_path is None:
				raise ValueError("Dataset path is required for ImageNet_VID dataset")
			dataset = ImageNet_VID(dataset_path, split)
		case "Random":
			dataset = Random(seed=442)
		case _:
			raise NotImplementedError(f"Dataset {dataset_name} not implemented")
	return dataset


@app.command()
def ls(
	repository: RepositoryPathType = Path("."),
	verbose: VerboseType = [],
):
	"""
	List all available models in the repository.
	"""
	metadata_list = asyncio.run(get_local_metadata(repository))
	print(
		"Models metadata parsed:\n\t" + "\n\t".join(m.name for m in metadata_list),
		flush=True,
	)


@app.command()
def check(
	model_name: ModelNameTypeOptional = None,
	weights_path: WeightsPathTypeOptional = None,
	skip_dependencies: Annotated[
		bool,
		typer.Option(help="If set to True, don't check dependencies."),
	] = False,
	# reinstall_dependencies: Annotated[
	# bool,
	# typer.Option(
	# help="If set to True, create a new environment and install dependencies for each model."
	# ),
	# ] = False,
	pin_installed: Annotated[
		bool,
		typer.Option(
			help=(
				"Set to True to pin dependencies to specific version into `requirements.txt`. "
				"Uses currently installed dependencies... its pip freeze equivalent"
			)
		),
	] = False,
	pin_defined: Annotated[
		bool,
		typer.Option(
			help=(
				"Set to True to pin dependencies to specific version into `requirements.txt`. "
				"If the repository contains a lock file they will be converted to requirements.txt"
				"Supported dependency sources: "
				"requirements.in, setup.cfg, setup.py, pyproject.toml, requirements.txt; "
				"Supported lock files: "
				"pipFile.lock, poetry.lock"
			)
		),
	] = False,
	repository: RepositoryPathType = Path("."),
	verbose: VerboseType = [],
):
	"""
	Check if model repository is correctly set up.

	It runs the following steps:

	1. Check if installed dependencies are pinned.
		- Skip with `--skip-dependencies`

	2. Parse metadata and select model (`MODEL_NAME`)

	3. Run inference on random data.
		- Only run if model name and weights are provided (`MODEL_NAME` and `WEIGHTS_PATH`)
	"""
	has_missing_dependencies = False
	if not skip_dependencies:
		has_missing_dependencies = asyncio.run(
			check_missing_dependencies(
				Path.cwd(), pin_installed=pin_installed, pin_defined=pin_defined
			)
		)
		logger.info("Dependencies checked!")

	metadata_list = asyncio.run(get_local_metadata(repository))
	logger.info("Model metadata successfully parsed")

	if model_name is not None:
		assert weights_path is not None
		metadata = next(
			x for x in metadata_list if x.name == model_name if model_name is not None
		)
		logger.info(f"Selected model: {metadata.name}")

		# if reinstall_dependencies:
		# with TemporaryDirectory() as venv:
		# raise NotImplementedError
		# runner.inference(Random(metadata.image_input))
		# else:
		async def run_inference():
			async with get_runner(metadata, weights_path) as runner:
				await runner.inference(Random(metadata.image_input).take(2))

		asyncio.run(run_inference())

	exit_code = 0
	if has_missing_dependencies:
		exit_code |= 8
	raise typer.Exit(code=exit_code)


@app.command()
def convert(
	weights_path: WeightsPathType,
	model_name: ModelNameType,
	model_format: Annotated[
		ModelFormat, typer.Argument(case_sensitive=False, help="Target model format")
	],
	quantization: Annotated[
		Quantization,
		typer.Option(
			"--quantization",
			"--quantize",
			case_sensitive=False,
			help="Quantization of the target model",
		),
	] = Quantization.none,
	hw_acceleration: Annotated[
		HardwareAcceleration,
		typer.Option(
			case_sensitive=False,
			help="Hardware acceleration of the target model. It is used only by Executorch",
		),
	] = HardwareAcceleration.CPU,
	override: Annotated[
		bool,
		typer.Option(
			help="Overwrite already converted weights if they exist (including intermediate ones)."
		),
	] = False,
	# calibration for quantization
	calibration_dataset: Annotated[
		str | None,
		typer.Option(
			"--calibration-dataset",
			help="Dataset name used to calibrate INT4/INT8 conversions. Required for INT4/INT8 conversions.",
		),
	] = None,
	calibration_dataset_path: Annotated[
		Path | None,
		typer.Option(
			"--calibration-dataset-path",
			help="Path with assets for the calibration dataset (when required).",
		),
	] = None,
	# verification
	verify: Annotated[
		bool,
		typer.Option(
			help="Run inference before/after conversion and compare.",
		),
	] = False,
	verify_dataset: Annotated[
		str | None,
		typer.Option(
			"--verify-dataset",
			help="Dataset name used for verification. Automatically sets `--verify` to True.",
			show_default="None; Random if `--verify` is set",
		),
	] = None,
	verify_dataset_path: Annotated[
		Path | None,
		typer.Option(
			"--verify-dataset-path",
			help="Path with assets for the verification dataset.",
		),
	] = None,
	# common
	artifacts: ArtifactsDirType = Path("build"),
	use_remote: SeparateEnvironmentType = False,
	repository_path: RepositoryPathType = Path("."),
	output: OutputType = Path("."),
	verbose: VerboseType = [],
):
	"""
	Convert an existing model to a different format. Target format can be quantized.
	"""
	assert quantization not in NEED_CALIBRATION or calibration_dataset is not None

	artifacts.mkdir(parents=True, exist_ok=True)
	output.mkdir(parents=True, exist_ok=True)

	metadata_list = asyncio.run(get_local_metadata(repository_path))
	metadata = next(x for x in metadata_list if x.name == model_name)

	if calibration_dataset is not None:
		parsed_calibration_dataset = (
			_get_dataset(
				calibration_dataset, calibration_dataset_path, split=DatasetSplit.val
			)
			.shuffle(42)
			.take(1000)
		)
	else:
		parsed_calibration_dataset = None

	if verify or verify_dataset is not None:
		parsed_verify_dataset = (
			_get_dataset(
				verify_dataset or "Random", verify_dataset_path, split=DatasetSplit.val
			)
			.shuffle(42)
			.take(1000)
		)
	else:
		parsed_verify_dataset = None

	try:
		results = asyncio.run(
			Convert(
				metadata.format,
				ExactModelFormat(model_format, quantization, hw_acceleration),
				calibration_dataset=parsed_calibration_dataset,
				verify_dataset=parsed_verify_dataset,
				override=override,
				use_remote=use_remote,
				error_immediately=True,
				metadata=metadata,
				output=output,
				artifacts=artifacts,
				default_model=weights_path,
			)()
		)
		if results is not None:
			print(results)
	except RemoteError as e:
		handle_remote_error(e)


@app.command()
def inference(
	weights_path: WeightsPathType,
	model_name: ModelNameTypeOptional = None,
	dataset_name: Annotated[
		str,
		typer.Argument(help="Dataset used for inference."),
	] = "Random",
	dataset_path: Annotated[
		Path | None,
		typer.Option(
			help="Optional path to the dataset assets (required for non-random datasets)."
		),
	] = None,
	take: Annotated[
		int,
		typer.Option(
			help="Number of samples to run through the model during inference.",
		),
	] = 10,
	split: Annotated[
		DatasetSplit, typer.Option(help="Dataset split to use.")
	] = DatasetSplit.val,
	# common
	artifacts: ArtifactsDirType = Path("build"),
	output: OutputTypeOptional = None,
	use_remote: SeparateEnvironmentType = False,
	repository_path: RepositoryPathType = Path("."),
	verbose: VerboseType = [],
):
	"""
	Run inference on a model using a selected dataset and print results.
	"""
	metadata_list = asyncio.run(get_local_metadata(repository_path))
	if model_name:
		metadata = next(x for x in metadata_list if x.name == model_name)
	else:
		assert weights_path.suffix == ".json", (
			f"Model name is required unless using a conversion result (weights_path: {weights_path})."
		)
		metadata = metadata_list[0]

	dataset = _get_dataset(dataset_name, dataset_path, split).take(take)
	try:
		results = asyncio.run(
			run_inference(
				weights_path,
				metadata,
				dataset,
				artifacts,
				output,
				{},
				use_remote,
			)
		)

		pprint(results)

	except RemoteError as e:
		handle_remote_error(e)


@app.callback(invoke_without_command=True)
def main(
	ctx: typer.Context,
	version: Annotated[
		bool, typer.Option("--version", "-v", help="Print version")
	] = False,
):
	if version:
		import importlib.metadata

		version_string_of_foo = importlib.metadata.version("ai-parade")
		print(version_string_of_foo)
		raise typer.Exit()
	if ctx.invoked_subcommand is None:
		pass


if __name__ == "__main__":
	app()



import numpy as np
import pytest

from ai_parade._toolkit.data import preprocessing


@pytest.mark.parametrize(
	"source_shape, target_shape",
	[
		((10, 10, 3), (10, 10)),
		((11, 11, 3), (10, 10)),
		((9, 9, 3), (10, 10)),
	],
)
def test_resize_resize_valid_shapes(
	source_shape: tuple[int, int, int], target_shape: tuple[int, int]
):
	source_image = np.zeros(source_shape, dtype=np.uint8)

	resized = preprocessing.resize(source_image, target_shape)

	assert resized.shape[:2] == target_shape


@pytest.mark.parametrize(
	"source_shape, target_shape",
	[
		((10, 10, 3), (10, 10)),
		((11, 11, 3), (10, 10)),
		((9, 9, 3), (10, 10)),
	],
)
def test_resize_resize_valid_shapes_with_padding_doing_nothing(
	source_shape: tuple[int, int, int], target_shape: tuple[int, int]
):
	source_image = np.zeros(source_shape, dtype=np.uint8)

	resized = preprocessing.resize(source_image, target_shape, padding=(1, 1, 1))

	assert resized.shape[:2] == target_shape
	assert resized.sum() == 0


@pytest.mark.parametrize(
	"source_shape, target_shape, padding_sum",
	[
		((10, 8, 3), (10, 10), 10 * 2 * 3),
		((11, 8, 3), (10, 10), 10 * 3 * 3),
		((9, 8, 3), (10, 10), 10 * 2 * 3),
	],
)
def test_resize_resize_valid_shapes_with_padding(
	source_shape: tuple[int, int, int], target_shape: tuple[int, int], padding_sum: int
):
	source_image = np.zeros(source_shape, dtype=np.uint8)

	resized = preprocessing.resize(source_image, target_shape, padding=(1, 1, 1))

	assert resized.shape[:2] == target_shape
	assert resized.sum() == padding_sum


def test_resize_resize_not_uniform_padding():
	source_image = np.zeros((10, 1, 3), dtype=np.uint8)

	resized = preprocessing.resize(source_image, (10, 10), padding=(1, 2, 3))

	assert resized.shape[:2] == (10, 10)
	assert resized.sum() == 10 * 9 * (1 + 2 + 3)
